class DiscodoException(Exception):
    pass


class Forbidden(DiscodoException):
    pass


class TooManyRequests(DiscodoException):
    pass


class NoSearchResults(DiscodoException):
    pass


class WebsocketConnectionClosed(DiscodoException):
    def __init__(self, socket, code=None) -> None:
        self.code = code or socket.close_code

        super().__init__(f"Websocket connection closed with {self.code}")


class VoiceClientNotFound(DiscodoException):
    pass


class NotConnected(DiscodoException):
    pass


class NotSeekable(DiscodoException):
    pass


class NotPlaying(DiscodoException):
    pass


class NodeNotConnected(DiscodoException):
    pass



import logging
import random
import re
from typing import Union

from youtube_related import RateLimited
from youtube_related import preventDuplication as relatedClient

from .config import Config
from .connector import VoiceConnector
from .errors import NotPlaying
from .player import Player
from .source import AudioData, AudioSource
from .utils import EventDispatcher

log = logging.getLogger("discodo.VoiceClient")

YOUTUBE_VIDEO_REGEX = re.compile(
    r"^((?:https?:)?\/\/)?((?:www|m)\.)?((?:youtube\.com|youtu.be))(\/(?:[\w\-]+\?v=|embed\/|v\/)?)([\w\-]+)(\S+)?$"
)


class VoiceClient(VoiceConnector):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.relatedClient = relatedClient()

        self.dispatcher = EventDispatcher()
        self.dispatcher.onAny(self.__dispatchToManager)
        self.dispatcher.on("REQUIRE_NEXT_SOURCE", self.__fetchAutoPlay)

        self.Queue = []
        self.player = None

        self._filter = {}
        self.paused = False

        self._autoplay = Config.DEFAULT_AUTOPLAY
        self._crossfade = Config.DEFAULT_CROSSFADE

        self._volume = Config.DEFAULT_VOLUME

        self.dispatcher.dispatch("VC_CREATED", id=self.id)

    def __del__(self) -> None:
        guild_id = int(self.guild_id) if self.guild_id else None

        log.info(f"destroying voice client of {guild_id}.")

        if self.manager.voiceClients.get(guild_id) == self:
            self.dispatcher.dispatch("VC_DESTROYED")
            del self.manager.voiceClients[guild_id]

        super().__del__()

        if self.player and self.player.is_alive():
            self.player.stop()

        for Item in self.Queue:
            if isinstance(Item, AudioSource):
                self.loop.call_soon_threadsafe(Item.cleanup)

    def __repr__(self) -> str:
        return f"<VoiceClient guild_id={self.guild_id} volume={self.volume} crossfade={self.crossfade} autoplay={self.autoplay}>"

    def __dispatchToManager(self, event, *args, **kwargs) -> None:
        self.manager.dispatcher.dispatch(self.guild_id, *args, event=event, **kwargs)

    async def __fetchAutoPlay(self, current, **_):
        for _ in range(5):
            if self.autoplay and not self.Queue:
                if YOUTUBE_VIDEO_REGEX.match(current.webpage_url):
                    address = Config.RoutePlanner.get() if Config.RoutePlanner else None
                    try:
                        Related = await self.relatedClient.async_get(
                            current.webpage_url, local_addr=address
                        )
                    except RateLimited:
                        Config.RoutePlanner.mark_failed_address(address)
                    else:
                        return await self.loadSource(
                            "https://www.youtube.com/watch?v=" + Related["id"],
                            related=True,
                        )

    def __spawnPlayer(self) -> None:
        if self.player and self.player.is_alive():
            return

        self.player = Player(self)
        self.player.crossfade = self.crossfade

        self.player.start()

    @property
    def channel_id(self) -> int:
        return self._channel_id

    @channel_id.setter
    def channel_id(self, value: int) -> None:
        self._channel_id = int(value)

        self.dispatcher.dispatch("VC_CHANNEL_EDITED", channel_id=self._channel_id)

    async def createSocket(self, data: dict = None) -> None:
        """
        Create UDP socket with discord to send voice packet.

        :param dict data: A VOICE_SERVER_UPDATE payload, defaults to None
        :rtype: None
        """

        await super().createSocket(data)

        self.__spawnPlayer()

    @property
    def state(self) -> str:
        """
        The current state of playback of the voice client

        :getter: Returns VoiceClient state
        :rtype: str
        """

        if not self.Queue and not self.player.current:
            return "stopped"
        elif self.paused:
            return "paused"
        else:
            return "playing"

    @property
    def volume(self) -> float:
        """
        Current volume of the voice client (0.0~2.0)

        :getter: Returns VoiceClient volume
        :setter: Set VoiceClient volume
        :rtype: float
        """

        return self._volume

    @volume.setter
    def volume(self, value: float) -> None:
        self._volume = round(max(value, 0.0), 2)

    @property
    def crossfade(self) -> float:
        """
        Current crossfade value of the voice client

        :getter: Returns VoiceClient crossfade
        :setter: Set VoiceClient crossfade
        :rtype: float
        """

        return self._crossfade

    @crossfade.setter
    def crossfade(self, value: float) -> None:
        if not isinstance(value, float):
            return TypeError("`crossfade` property must be `float`.")

        self.player.crossfade = self._crossfade = round(max(value, 0.0), 1)

    @property
    def autoplay(self) -> bool:
        """
        Current autoplay value of the voice client

        :getter: Returns VoiceClient autoplay
        :setter: Set VoiceClient autoplay
        :rtype: bool
        """

        return self._autoplay

    @autoplay.setter
    def autoplay(self, value: bool) -> None:
        if not isinstance(value, bool):
            return TypeError("`autoplay` property must be `bool`.")

        self._autoplay = value

    @property
    def filter(self) -> dict:
        """
        Current filter value of the voice client

        :getter: Returns VoiceClient filter
        :setter: Set VoiceClient filter
        :rtype: dict
        """

        return self._filter

    @filter.setter
    def filter(self, value: dict) -> None:
        if not isinstance(value, dict):
            return TypeError("`filter` property must be `dict`.")

        self._filter = value

    @property
    def current(self) -> Union[AudioSource, AudioData]:
        """
        Current source of the voice client

        :getter: Returns current source
        :rtype: :class:`AudioSource` or :class:`AudioData`
        """

        return self.player.current

    @property
    def next(self) -> Union[AudioSource, AudioData]:
        """
        Next source of the voice client

        :getter: Returns next source
        :rtype: :class:`AudioSource` or :class:`AudioData`
        """

        return self.player.next

    def putSource(self, Source: Union[list, AudioData, AudioSource]) -> int:
        """
        Put source to Queue.

        :param Source: list or :class:`AudioSource` or :class:`AudioData` to put on Queue
        :rtype: int
        """

        if not isinstance(Source, (list, AudioData, AudioSource)):
            raise TypeError("`Source` must be `list` or `AudioData` or `AudioSource`.")

        self.Queue += Source if isinstance(Source, list) else [Source]

        self.dispatcher.dispatch(
            "putSource", sources=(Source if isinstance(Source, list) else [Source])
        )

        return (
            self.Queue.index(Source)
            if not isinstance(Source, list)
            else [self.Queue.index(Item) for Item in Source]
        )

    async def getSource(self, Query: str) -> AudioData:
        """
        Get source from Query

        :param str Query: Query to search
        :rtype: :class:`AudioData`
        """

        return await AudioData.create(Query)

    async def loadSource(self, Query: str, **kwargs) -> AudioData:
        """
        Get source from Query and put it on Queue.

        :param str Query: Query to search
        :rtype: :class:`AudioData`
        """

        Data = await self.getSource(Query)

        if "related" in kwargs and isinstance(kwargs["related"], bool):
            Data.related = kwargs["related"]

        Index = self.putSource(Data)

        self.dispatcher.dispatch(
            "loadSource",
            source=(
                {"data": Data, "index": Index}
                if not isinstance(Data, list)
                else list(
                    map(
                        lambda zipped: {"data": zipped[0], "index": zipped[1]},
                        zip(Data, Index),
                    )
                )
            ),
            **kwargs,
        )

        return Data

    async def seek(self, offset: int) -> None:
        """
        Seek current playing source.

        :param float offset: offset to seek
        :rtype: None
        """

        return await self.player.seek(offset)

    def skip(self, offset: int = 1) -> None:
        """
        skip current playing source.

        :param float offset: offset to skip
        :rtype: None
        """

        if not self.player.current:
            raise NotPlaying

        if len(self.Queue) < (offset - 1):
            raise ValueError("`offset` is bigger than `Queue` size.")

        if offset > 1:
            del self.Queue[0 : (offset - 1)]

        self.player.current.skip()

    def pause(self) -> bool:
        """
        Pause playing

        :rtype: True
        """

        self.paused = True
        return True

    def resume(self) -> bool:
        """
        Resume playing

        :rtype: True
        """

        self.paused = False
        return False

    def shuffle(self) -> dict:
        """
        Shuffle Queue.

        :rtype: dict
        """

        if not self.Queue:
            raise ValueError("`Queue` is empty now.")

        random.shuffle(self.Queue)

        return self.Queue



import re

YOUTUBE_HEADERS: dict = {
    "x-youtube-client-name": "1",
    "x-youtube-client-version": "2.20201030.01.00",
}

DATA_JSON = re.compile(r'(?:window\["ytInitialData"\]|ytInitialData)\W?=\W?({.*?});')

from .mix import extract_mix
from .playlist import extract_playlist
from .search import search
from .subtitle import get_subtitle


class Youtube:
    extract_mix = extract_mix
    extract_playlist = extract_playlist
    search = search
    get_subtitle = get_subtitle



import re
from typing import Union

from asyncspotify import Client, ClientCredentialsFlow

from ...config import Config

URL_REGEX = re.compile(
    r"(?:http(?:s):\/\/open\.spotify\.com\/|spotify:)(playlist|track|album)(?:\/|:)([a-z|A-Z|0-9]+)"
)

spotify = None


async def create_client():
    global spotify

    if spotify:
        return

    import warnings

    warnings.warn(
        "Sporty features are experimental, do not use on production.", UserWarning
    )

    auth = ClientCredentialsFlow(
        client_id=Config.SPOTIFY_ID,
        client_secret=Config.SPOTIFY_SECRET,
    )

    spotify = Client(auth)

    await spotify.authorize()


async def get_track(id: str) -> str:
    Track = await spotify.get_track(id)

    return f"{Track.artists.pop().name} - {Track.name}"


async def get_album(id: str) -> list:
    Tracks = await spotify.get_album_tracks(id)

    return [f"{Track.artists.pop().name} - {Track.name}" for Track in Tracks]


async def get_playlist(id: str) -> list:
    Tracks = await spotify.get_playlist_tracks(id)

    return [f"{Track.artists.pop().name} - {Track.name}" for Track in Tracks]


async def resolve(query: str) -> Union[str, list]:
    if not Config.SPOTIFY_ID or not Config.SPOTIFY_SECRET:
        return query

    if not spotify:
        await create_client()

    URL_PARSE = URL_REGEX.search(query)
    if URL_PARSE:
        if URL_PARSE.group(1) == "track":
            return await get_track(URL_PARSE.group(2))
        elif URL_PARSE.group(1) == "album":
            return await get_album(URL_PARSE.group(2))
        elif URL_PARSE.group(1) == "playlist":
            return await get_playlist(URL_PARSE.group(2))

    return query



import asyncio
import ipaddress
import re
import urllib.parse
from typing import Union

import aiohttp

from ..errors import NoSearchResults
from .resolver import resolve
from .youtube import Youtube
from .youtube_dl import extract as youtube_dl_extract

URL_REGEX = re.compile(
    r"(?:https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|www\.[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9]+\.[^\s]{2,}|www\.[a-zA-Z0-9]+\.[^\s]{2,})"
)

YOUTUBE_PLAYLIST_ID_REGEX = re.compile(
    r"(?:http|https|)(?::\/\/|)(?:www.|)(?:music.|)(?:youtu\.be\/|youtube\.com(?:\/embed\/|\/v\/|\/watch\?v=|\/ytscreeningroom\?v=|\/feeds\/api\/videos\/|\/user\S*[^\w\-\s]|\S*[^\w\-\s]))([\w\-]{12,})[a-z0-9;:@#?&%=+\/\$_.-]*(?:&index=|)([0-9]*)?"
)


async def extract(
    query: Union[str, list],
    address: Union[ipaddress.IPv4Address, ipaddress.IPv6Address] = None,
    **kwargs
) -> dict:
    connector = aiohttp.TCPConnector(local_addr=(str(address), 0)) if address else None

    query = await resolve(query, connector)

    if isinstance(query, list):
        Done, _ = await asyncio.wait(
            [youtube_dl_extract(Item, address=address, **kwargs) for Item in query]
        )

        Results = []
        for Future in Done:
            try:
                Results.append(Future.result())
            except:
                pass

        return Results

    if not URL_REGEX.match(query):
        searchResult: list = await Youtube.search(query, connector)

        if not searchResult:
            raise NoSearchResults

        return searchResult[0]

    Match = YOUTUBE_PLAYLIST_ID_REGEX.match(query)
    if Match:
        if Match.group(1).startswith(("RD", "UL", "PU")):
            urlInfo = urllib.parse.parse_qs(urllib.parse.urlparse(query).query)

            if urlInfo.get("v") and urlInfo.get("list"):
                return await Youtube.extract_mix(
                    urlInfo["v"][0], urlInfo.get("list")[0], connector
                )
        else:
            return await Youtube.extract_playlist(Match.group(1), connector)

    return await youtube_dl_extract(query, address=address, **kwargs)


search = Youtube.search



from http.client import responses
from typing import Union

import aiohttp

from ..errors import DiscodoException


class HTTPException(DiscodoException):
    def __init__(self, status: int) -> None:
        super().__init__(f"{status} {responses.get(status, 'Unknown Status Code')}")


class HTTPClient:
    def __init__(self, client) -> None:
        self.VoiceClient = client
        self.Node = client.Node
        self.loop = client.Node.loop

    @property
    def headers(self) -> dict:
        return {
            "Authorization": self.Node.password,
            "User-ID": str(self.Node.user_id),
            "Guild-ID": str(self.VoiceClient.guild_id),
            "VoiceClient-ID": str(self.VoiceClient.id),
        }

    async def fetch(self, method: str, endpoint: str, **kwargs) -> dict:
        URL = self.Node.URL + endpoint

        async with aiohttp.ClientSession(headers=self.headers) as session:
            async with session.request(method, URL, **kwargs) as response:
                if 200 <= response.status < 300:
                    return await response.json(content_type=None)

                raise HTTPException(response.status)

    async def getSource(self) -> dict:
        return (await self.fetch("GET", "/getSource"))["source"]

    async def searchSource(self) -> list:
        return (await self.fetch("GET", "/searchSource"))["sources"]

    async def putSource(self, source: dict) -> Union[int, list]:
        return (await self.fetch("POST", "/putSource", json={"source": dict(source)}))[
            "index"
        ]

    async def loadSource(self, query: str) -> dict:
        return (await self.fetch("POST", "/loadSource", json={"query": str(query)}))[
            "source"
        ]

    async def setVolume(self, volume: float) -> None:
        return await self.fetch("POST", "/setVolume", json={"volume": float(volume)})

    async def setCrossfade(self, crossfade: float) -> None:
        return await self.fetch(
            "POST", "/setCrossfade", json={"crossfade": float(crossfade)}
        )

    async def setAutoplay(self, autoplay: bool) -> None:
        return await self.fetch(
            "POST", "/setAutoplay", json={"autoplay": bool(autoplay)}
        )

    async def setFilter(self, filter: dict) -> None:
        return await self.fetch("POST", "/setFilter", json={"filter": dict(filter)})

    async def seek(self, offset: float) -> None:
        return await self.fetch("POST", "/seek", json={"offset": float(offset)})

    async def skip(self, offset: int) -> None:
        return (await self.fetch("POST", "/skip", json={"offset": int(offset)}))[
            "remain"
        ]

    async def pause(self) -> None:
        return await self.fetch("POST", "/pause")

    async def resume(self) -> None:
        return await self.fetch("POST", "/pause")

    async def shuffle(self) -> None:
        return await self.fetch("POST", "/shuffle")

    async def remove(self, index: int) -> None:
        return await self.fetch("POST", "/remove", json={"index": int(index)})

    async def getState(self) -> dict:
        return await self.fetch("GET", "/state")

    async def getQueue(self) -> dict:
        return (await self.fetch("GET", "/queue"))["entries"]



import asyncio

from ..utils import EventDispatcher
from .http import HTTPClient


class VoiceClient:
    def __init__(self, Node, id: str, guild_id: int) -> None:
        self.Node = Node
        self.loop = Node.loop

        self.id = id
        self.guild_id = guild_id
        self.channel_id: int = None

        self.http = HTTPClient(self)
        self.dispatcher = EventDispatcher()

        self.dispatcher.on("VC_CHANNEL_EDITED", self.__channel_editied)

    def __repr__(self) -> str:
        return f"<VoiceClient id={self.id} guild_id={self.guild_id} channel_id={self.channel_id} Node={self.Node}>"

    def __del__(self):
        vc = self.Node.voiceClients.get(self.guild_id)
        if vc and vc == self:
            self.Node.voiceClients.pop(self.guild_id)

    def __channel_editied(self, channel_id: int) -> None:
        self.channel_id: int = channel_id

    async def send(self, Operation: str, Data: dict = {}):
        Data["guild_id"] = self.guild_id

        return await self.Node.send(Operation, Data)

    async def query(
        self, Operation: str, Data: dict = {}, Event: str = None, timeout: float = 10.0
    ) -> dict:
        if not Event:
            Event = Operation

        Future = self.loop.create_task(
            self.dispatcher.wait_for(
                Event,
                condition=lambda Data: int(Data["guild_id"]) == int(self.guild_id),
                timeout=timeout,
            )
        )

        await self.send(Operation, Data)

        return await Future

    async def getSource(self, Query: str, ws: bool = True) -> dict:
        if ws:
            return (await self.query("getSource", {"query": Query}))["source"]

        return await self.http.getSource(Query)

    async def searchSources(self, Query: str, ws: bool = True) -> list:
        if ws:
            return (await self.query("searchSources", {"query": Query}))["sources"]

        return await self.http.searchSources(Query)

    async def loadSource(self, Query: str, ws: bool = True) -> dict:
        if ws:
            return (await self.query("loadSource", {"query": Query}))["source"]

        return await self.http.loadSource(Query)

    async def putSource(self, Source: dict, ws: bool = True) -> int:
        if ws:
            return (await self.query("putSource", {"song": Source}))["index"]

        return await self.http.putSource(Source)

    async def skip(self, offset: int = 1, ws: bool = True) -> int:
        if ws:
            return (await self.query("skip", {"offset": offset}))["remain"]

        return await self.http.skip(offset)

    async def seek(self, offset: float, ws: bool = True) -> dict:
        if ws:
            return await self.query("seek", {"offset": offset})

        return await self.http.seek(offset)

    async def setVolume(self, volume: int, ws: bool = True) -> float:
        if ws:
            return (await self.query("setVolume", {"volume": volume}))["volume"]

        return await self.http.setVolume(volume)

    async def setCrossfade(self, crossfade: float, ws: bool = True) -> float:
        if ws:
            return (await self.query("setCrossfade", {"crossfade": crossfade}))[
                "crossfade"
            ]

        return await self.http.setCrossfade(crossfade)

    async def setAutoplay(self, autoplay: bool, ws: bool = True) -> bool:
        if ws:
            return (await self.query("setAutoplay", {"autoplay": autoplay}))["autoplay"]

        return await self.http.setAutoplay(autoplay)

    async def setFilter(self, filter: dict, ws: bool = True) -> dict:
        if ws:
            return await self.query("setFilter", {"filter": filter})

        return await self.http.setFilter(filter)

    async def pause(self, ws: bool = True) -> dict:
        if ws:
            return await self.query("pause")

        return await self.http.pause()

    async def resume(self, ws: bool = True) -> dict:
        if ws:
            return await self.query("resume")

        return await self.http.resume()

    async def getQueue(self, ws: bool = True) -> list:
        if ws:
            return (await self.query("getQueue"))["entries"]

        return await self.http.getQueue()

    async def getState(self, ws: bool = True) -> dict:
        if ws:
            return await self.query("getState")

        return await self.http.getState()

    async def shuffle(self, ws: bool = True) -> dict:
        if ws:
            return await self.query("shuffle")

        return await self.http.shuffle()

    async def remove(self, index: int, ws: bool = True) -> dict:
        if ws:
            return await self.query("remove", {"index": index})

        return await self.http.remove(index)

    async def requestSubtitle(self, lang: str = None, url: str = None) -> dict:
        if not any([lang, url]):
            raise ValueError("Either `lang` or `url` is needed.")

        Data = {}
        if url:
            Data["url"] = url
        elif lang:
            Data["lang"] = lang

        return await self.query("requestSubtitle", Data)

    async def getSubtitle(self, *args, callback: callable, **kwargs):
        if not asyncio.iscoroutinefunction(callback):
            raise ValueError("Callback function must be coroutine function.")

        Data = await self.requestSubtitle(*args, **kwargs)

        identify_token = Data.get("identify")
        if not identify_token:
            raise ValueError(f"Subtitle not found.")

        _lyricsLock = asyncio.Lock()

        async def lyricsRecieve(lyrics):
            if lyrics["identify"] != identify_token or _lyricsLock.locked():
                return

            await _lyricsLock.acquire()

            try:
                await callback(lyrics)
            finally:
                _lyricsLock.release()

        async def lyricsDone(Data):
            if Data["identify"] != identify_token:
                return

            self.dispatcher.off("Subtitle", lyricsRecieve)
            self.dispatcher.off("subtitleDone", lyricsDone)

        self.dispatcher.on("Subtitle", lyricsRecieve)
        self.dispatcher.on("subtitleDone", lyricsDone)

        return Data

    async def destroy(self) -> dict:
        return await self.query("VC_DESTROY", Event="VC_DESTROYED")



import re
from typing import Union

import aiohttp
from bs4 import BeautifulSoup

MELON_REGEX = re.compile(
    r"(?:https?:\/\/)?(?:www\.)?melon.com\/album\/detail.htm\?albumId=([0-9]+)"
)


async def get_album(url: str, connector: aiohttp.TCPConnector = None) -> list:
    async with aiohttp.ClientSession(
        headers={
            "User-Agent": "Mozilla/5.0 (Windows NT 6.3; Trident/7.0; rv:11.0) like Gecko"
        }
    ) as session:
        async with session.get(url, connector=connector) as resp:
            Body = await resp.text()

    soup = BeautifulSoup(Body, "html.parser")

    Table = soup.find("form", attrs={"id": "frm"}).find("table")

    Items = [
        Element.find_all("div", class_="ellipsis")
        for Element in [
            Item.find("div", class_="wrap_song_info") for Item in Table.find_all("tr")
        ]
        if Element
    ]

    return [
        f'{Item[1].find("a").text} - {Item[0].find("a").text}' for Item in Items if Item
    ]


async def resolve(
    query: str, connector: aiohttp.TCPConnector = None
) -> Union[str, list]:
    is_album = MELON_REGEX.match(query)

    if is_album:
        Album = await get_album(query, connector)

        if len(Album) == 1:
            return Album[0]

        return Album

    return query



import logging
import urllib.parse
from xml.etree import ElementTree

import aiohttp

log = logging.getLogger("discodo.extractor.youtube")


async def get_subtitle(videoId: str, connector: aiohttp.TCPConnector = None) -> list:
    log.info(f"Downloading subtitle page of {videoId}")
    async with aiohttp.ClientSession(connector=connector) as session:
        async with session.get(
            "https://video.google.com/timedtext",
            params={"hl": "en", "type": "list", "v": videoId},
        ) as resp:
            Body: str = await resp.text()

    Tree = ElementTree.fromstring(Body)

    Subtitles: dict = {}

    for Track in Tree.findall("track"):
        if Track.attrib["lang_code"] in Subtitles:
            continue

        Params = urllib.parse.urlencode(
            {
                "lang": Track.attrib["lang_code"],
                "v": videoId,
                "fmt": "srv1",
                "name": Track.attrib["name"].encode("utf-8"),
            }
        )

        Subtitles[Track.attrib["lang_code"]] = {
            "url": "https://www.youtube.com/api/timedtext?" + Params,
            "ext": "srv1",
        }

    return Subtitles



from typing import Any, Coroutine

from ..errors import NotSeekable
from .PyAVSource import PyAVSource


class AudioSource(PyAVSource):
    def __init__(self, *args, AudioData=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.AudioData = AudioData
        self.address = self.AudioData.address if self.AudioData else None

        self._skipped: bool = False

    def __dict__(self) -> dict:
        Value = self.AudioData.__dict__() if self.AudioData else {}

        Value["_type"] = "AudioSource"
        Value["seekable"] = self.seekable
        Value["duration"] = self.duration
        if hasattr(self, "position"):
            Value["position"] = self.position

        return Value

    def __repr__(self) -> str:
        return (
            f"<AudioSource id={self.id} title='{self.title}' duration={self.duration}"
            + (f" position={self.position}" if hasattr(self, "position") else "")
            + f" seekable={self.seekable} address='{self.address}'>"
        )

    def __getattr__(self, key: str) -> Any:
        return getattr(self.AudioData, key)

    @property
    def seekable(self) -> bool:
        return not self.AudioData.is_live if self.AudioData else True

    def seek(self, *args, **kwargs) -> Coroutine:
        if not self.seekable:
            raise NotSeekable

        return super().seek(*args, **kwargs)

    @property
    def duration(self) -> float:
        return self._duration if self._duration else self.AudioData.duration

    @property
    def remain(self) -> float:
        return round(self.duration - self.position, 2)

    @property
    def filter(self) -> dict:
        return self._filter

    @filter.setter
    def filter(self, value: dict) -> dict:
        if self.AudioData and self.AudioData.is_live and "atempo" in value:
            raise ValueError("Cannot use `atempo` filter in live streaming.")

        self._filter = value

    @property
    def skipped(self) -> bool:
        return self._skipped

    @skipped.setter
    def skipped(self, Value: bool) -> None:
        if Value:
            self.stop()

        self._skipped = Value

    def skip(self) -> None:
        self.skipped = True



import json
import logging

import aiohttp

from . import DATA_JSON, YOUTUBE_HEADERS

log = logging.getLogger("discodo.extractor.youtube")


async def search(Query: str, connector: aiohttp.TCPConnector = None) -> None:
    log.info(f"Downloading search page of {Query}")
    async with aiohttp.ClientSession(
        headers=YOUTUBE_HEADERS, connector=connector
    ) as session:
        async with session.get(
            "https://www.youtube.com/results", params={"search_query": Query}
        ) as resp:
            Body: str = await resp.text()

    Search = DATA_JSON.search(Body)

    if not Search:
        raise ValueError

    Data = json.loads(Search.group(1))

    def extract_video(Data: dict) -> dict:
        Renderer: dict = Data.get("videoRenderer")

        if not Renderer or not Renderer.get("lengthText"):
            return

        return {
            "id": Renderer["videoId"],
            "title": Renderer["title"]["runs"][0]["text"],
            "webpage_url": "https://www.youtube.com/watch?v=" + Renderer["videoId"],
            "uploader": Renderer["ownerText"]["runs"][0]["text"],
            "duration": sum(
                [
                    int(value) * (60 ** index)
                    for index, value in enumerate(
                        reversed(Renderer["lengthText"]["simpleText"].split(":"))
                    )
                ]
            ),
        }

    Videos: list = list(
        filter(
            None,
            map(
                extract_video,
                Data["contents"]["twoColumnSearchResultsRenderer"]["primaryContents"][
                    "sectionListRenderer"
                ]["contents"][0]["itemSectionRenderer"]["contents"],
            ),
        )
    )

    return Videos



import json
import logging

import aiohttp

from . import DATA_JSON, YOUTUBE_HEADERS

log = logging.getLogger("discodo.extractor.youtube")


async def extract_mix(
    vId: str, playlistId: str, connector: aiohttp.TCPConnector = None
) -> dict:
    if not playlistId.startswith(("RD", "UL", "PU")):
        raise TypeError("playlistId is not Youtube Mix id")

    log.info(f"Downloading playlist page of {playlistId}")
    async with aiohttp.ClientSession(
        headers=YOUTUBE_HEADERS, connector=connector
    ) as session:
        async with session.get(
            f"https://www.youtube.com/watch", params={"v": vId, "list": playlistId}
        ) as resp:
            Body: str = await resp.text()

    Search = DATA_JSON.search(Body)

    if not Search:
        raise ValueError

    Data: dict = json.loads(Search.group(1))
    Playlist: dict = Data["contents"]["twoColumnWatchNextResults"]["playlist"][
        "playlist"
    ]

    def extract(Track: dict) -> dict:
        Renderer: dict = Track["playlistPanelVideoRenderer"]

        return {
            "id": Renderer["videoId"],
            "title": Renderer["title"]["simpleText"],
            "webpage_url": "https://youtube.com/watch?v=" + Renderer["videoId"],
            "uploader": Renderer["longBylineText"]["runs"][0]["text"],
            "duration": sum(
                [
                    int(value) * (60 ** index)
                    for index, value in enumerate(
                        reversed(Renderer["lengthText"]["simpleText"].split(":"))
                    )
                ]
            ),
        }

    return list(map(extract, Playlist["contents"]))



import asyncio
import concurrent.futures
import json
import logging
import struct
import threading
import time
from collections import deque
from typing import Dict, Optional

import aiohttp

from .errors import WebsocketConnectionClosed
from .natives import Cipher

log = logging.getLogger("discodo.gateway")


class keepAliver(threading.Thread):
    def __init__(self, ws, interval: int, *args, **kwargs) -> None:
        threading.Thread.__init__(self, *args, **kwargs)
        self.daemon = True

        self.ws = ws
        self.interval: int = interval
        self.Stopped = threading.Event()
        self.latency: Optional[float] = None
        self.recent_latencies = deque(maxlen=20)

        self._lastAck = self._lastSend = time.perf_counter()
        self.timeout: float = ws.heartbeatTimeout
        self.threadId: int = ws.threadId

    def __del__(self) -> None:
        self.stop()

    def run(self) -> None:
        while not self.Stopped.wait(self.interval):
            if (self._lastAck + self.timeout) < time.perf_counter():
                Runner = asyncio.run_coroutine_threadsafe(
                    self.ws.close(4000), self.ws.loop
                )

                try:
                    Runner.result()
                except:
                    pass
                finally:
                    return self.stop()

            payload: Dict[str, int] = {
                "op": VoicePayload.HEARTBEAT,
                "d": int(time.time() * 1000),
            }
            Runner = asyncio.run_coroutine_threadsafe(
                self.ws.sendJson(payload), self.ws.loop
            )
            try:
                totalBlocked: int = 0
                while True:
                    try:
                        Runner.result(10)
                        break
                    except concurrent.futures.TimeoutError:
                        totalBlocked += 10
                        log.warning(
                            f"Thread-{self.threadId}: Heartbeat blocked for more than {totalBlocked} seconds."
                        )
            except:
                return self.stop()
            else:
                self._lastSend = time.perf_counter()

    def ack(self) -> None:
        self._lastAck: float = time.perf_counter()
        self.latency: float = self._lastAck - self._lastSend
        self.recent_latencies.append(self.latency)

    def stop(self) -> None:
        self.Stopped.set()


class VoicePayload:
    IDENTIFY = 0
    SELECT_PROTOCOL = 1
    READY = 2
    HEARTBEAT = 3
    SESSION_DESCRIPTION = 4
    SPEAKING = 5
    HEARTBEAT_ACK = 6
    RESUME = 7
    HELLO = 8
    RESUMED = 9
    CLIENT_DISCONNECT = 13


class VoiceSocket:
    def __init__(self, client, session, socket) -> None:
        self.loop = client.loop
        self.client = client
        self.session = session
        self.socket = socket

        self.closeCode = None

        self.keepAliver = None
        self.heartbeatTimeout = 60.0
        self.threadId = threading.get_ident()

    def __del__(self) -> None:
        self.loop.create_task(self.close())

    @classmethod
    async def connect(cls, client, resume=False):
        session = aiohttp.ClientSession()
        socket = await session.ws_connect(
            f"wss://{client.endpoint}/?v=4",
            max_msg_size=0,
            timeout=30.0,
            autoclose=False,
            compress=15,
        )

        ws = cls(client, session, socket)

        if not resume:
            await ws.identify()
        else:
            await ws.resume()

        return ws

    @property
    def latency(self) -> float:
        return self.keepAliver.latency if self.keepAliver else None

    @property
    def averageLatency(self) -> float:
        if not self.keepAliver:
            return None

        return sum(self.keepAliver.recent_latencies) / len(
            self.keepAliver.recent_latencies
        )

    async def sendJson(self, data: dict) -> None:
        log.debug(f"send to websocket {data}")
        await self.socket.send_json(data)

    async def identify(self) -> None:
        payload = {
            "op": VoicePayload.IDENTIFY,
            "d": {
                "server_id": str(self.client.guild_id),
                "user_id": str(self.client.user_id),
                "session_id": self.client.session_id,
                "token": self.client.token,
            },
        }
        await self.sendJson(payload)

    async def resume(self) -> None:
        payload = {
            "op": VoicePayload.RESUME,
            "d": {
                "server_id": str(self.client.guild_id),
                "user_id": str(self.client.user_id),
                "session_id": self.client.session_id,
                "token": self.client.token,
            },
        }
        await self.sendJson(payload)

    async def select_protocol(self, ip: str, port: str, mode: str) -> None:
        payload = {
            "op": VoicePayload.SELECT_PROTOCOL,
            "d": {
                "protocol": "udp",
                "data": {"address": ip, "port": port, "mode": mode},
            },
        }
        await self.sendJson(payload)

    async def speak(self, state: bool = True) -> None:
        self.client.speakState = state

        payload = {
            "op": VoicePayload.SPEAKING,
            "d": {"speaking": int(state), "delay": 0},
        }
        await self.sendJson(payload)

    async def messageRecieved(self, message: dict) -> None:
        Operation, Data = message["op"], message.get("d")
        log.debug(f"websocket recieved {Operation}: {Data}")

        if Operation == VoicePayload.READY:
            await self.createConnection(Data)
        elif Operation == VoicePayload.HEARTBEAT_ACK:
            self.keepAliver.ack()
        elif Operation == VoicePayload.SESSION_DESCRIPTION:
            await self.loadKey(Data)
        elif Operation == VoicePayload.HELLO:
            interval = Data["heartbeat_interval"] / 1000.0
            self.keepAliver = keepAliver(self, min(interval, 5.0))
            self.keepAliver.start()

    async def createConnection(self, data: dict) -> None:
        self.client.ssrc = data["ssrc"]
        self.client.endpointIp = data["ip"]
        self.client.endpointPort = data["port"]

        packet = bytearray(70)
        struct.pack_into(">I", packet, 0, data["ssrc"])

        self.client.socket.sendto(
            packet, (self.client.endpointIp, self.client.endpointPort)
        )
        _recieved = await self.loop.sock_recv(self.client.socket, 70)

        start, end = 4, _recieved.index(0, 4)
        self.client.ip = _recieved[start:end].decode("ascii")
        self.client.port = struct.unpack_from(">H", _recieved, len(_recieved) - 2)[0]

        encryptModes = [Mode for Mode in data["modes"] if Mode in Cipher.available]
        log.debug(f'recieved encrypt modes {data["modes"]}')

        encryptMode = encryptModes[0]
        log.info(f"select encrypt mode {encryptMode}")

        await self.select_protocol(self.client.ip, self.client.port, encryptMode)

    async def loadKey(self, data: dict) -> None:
        log.info("recieved voice secret key.")

        await self.speak(True)
        await self.speak(False)

        self.client.encryptMode = data["mode"]
        self.client.secretKey = data.get("secret_key")

    async def poll(self) -> None:
        message = await asyncio.wait_for(self.socket.receive(), timeout=30.0)

        if message.type is aiohttp.WSMsgType.TEXT:
            await self.messageRecieved(json.loads(message.data))
        elif message.type is aiohttp.WSMsgType.ERROR:
            raise WebsocketConnectionClosed(self.socket) from message.data
        elif message.type in (
            aiohttp.WSMsgType.CLOSED,
            aiohttp.WSMsgType.CLOSE,
            aiohttp.WSMsgType.CLOSING,
        ):
            raise WebsocketConnectionClosed(self.socket, code=self.socket.close_code)

    async def close(self, code: int = 1000) -> None:
        if self.keepAliver:
            self.keepAliver.stop()

        self._close_code = code
        await self.socket.close(code=code)
        await self.session.close()



import json
import logging

import aiohttp

from ...config import Config
from . import DATA_JSON, YOUTUBE_HEADERS

log = logging.getLogger("discodo.extractor.youtube")


async def extract_playlist(
    playlistId: str, connector: aiohttp.TCPConnector = None
) -> dict:
    if playlistId.startswith(("RD", "UL", "PU")):
        raise TypeError("playlistId is Youtube Mix id")

    log.info(f"Downloading playlist page of {playlistId}")
    async with aiohttp.ClientSession(
        headers=YOUTUBE_HEADERS, connector=connector
    ) as session:
        async with session.get(
            f"https://www.youtube.com/playlist", params={"list": playlistId, "hl": "en"}
        ) as resp:
            Body: str = await resp.text()

        Search = DATA_JSON.search(Body)

        if not Search:
            raise ValueError

        Data: dict = json.loads(Search.group(1))

        if Data.get("alerts"):
            raise Exception(Data["alerts"][0]["alertRenderer"]["text"]["simpleText"])

        firstPlaylistData: list = Data["contents"]["twoColumnBrowseResultsRenderer"][
            "tabs"
        ][0]["tabRenderer"]["content"]["sectionListRenderer"]["contents"][0][
            "itemSectionRenderer"
        ][
            "contents"
        ][
            0
        ][
            "playlistVideoListRenderer"
        ]

        Sources: list = []

        def extract_playlist(playlistData: dict, name: str = "contents") -> str:
            trackList: list = playlistData.get(name)
            if not trackList:
                return

            continuationsTokens: list = []

            def extract(Track: dict) -> dict:
                if "playlistVideoRenderer" in Track:
                    Renderer: dict = Track.get("playlistVideoRenderer", {})
                    shortBylineText: dict = Renderer.get("shortBylineText")

                    if not Renderer.get("isPlayable") or not shortBylineText:
                        return

                    return {
                        "id": Renderer["videoId"],
                        "title": Renderer["title"].get("simpleText")
                        or Renderer["title"]["runs"][0]["text"],
                        "webpage_url": "https://youtube.com/watch?v="
                        + Renderer["videoId"],
                        "uploader": shortBylineText["runs"][0]["text"],
                        "duration": Renderer["lengthSeconds"],
                    }
                elif "continuationItemRenderer" in Track:
                    continuationsTokens.append(
                        Track["continuationItemRenderer"]["continuationEndpoint"][
                            "continuationCommand"
                        ]["token"]
                    )

                    return
                else:
                    return

            Sources.extend(map(extract, trackList))

            if not continuationsTokens:
                return

            return (
                "https://www.youtube.com/browse_ajax?continuation="
                + continuationsTokens[0]
                + "&ctoken="
                + continuationsTokens[0]
                + "&hl=en"
            )

        continuations_url: str = extract_playlist(firstPlaylistData)
        for _ in range(Config.PLAYLIST_PAGE_LIMIT):
            if not continuations_url:
                break

            log.info(f"Downloading playlist continuation page of {playlistId}")
            async with session.get(continuations_url) as resp:
                Body: dict = await resp.json()

            nextPlaylistData: dict = Body[1]["response"]["onResponseReceivedActions"][
                0
            ]["appendContinuationItemsAction"]

            continuations_url: str = extract_playlist(
                nextPlaylistData, name="continuationItems"
            )

        return list(filter(None, Sources))



import logging
import threading

import av

from ..config import Config

log = logging.getLogger("discodo.natives.AudioFifo")


class AudioFifo(av.AudioFifo):
    SAMPLES_PER_FRAME = Config.SAMPLES_PER_FRAME

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.AUDIOBUFFERLIMITMS = Config.BUFFERLIMIT * 50 * self.SAMPLES_PER_FRAME

        self.haveToFillBuffer = threading.Event()
        self.haveToFillBuffer.set()

    def is_ready(self) -> bool:
        return self.samples >= self.SAMPLES_PER_FRAME

    def check_buffer(self) -> None:
        if self.samples < self.AUDIOBUFFERLIMITMS:
            self.haveToFillBuffer.set()
        else:
            self.haveToFillBuffer.clear()

    def read(self, samples: int = SAMPLES_PER_FRAME) -> bytes:
        AudioFrame = super().read(samples)
        if not AudioFrame:
            return

        self.check_buffer()

        return AudioFrame.planes[0].to_bytes()

    def write(self, *args, **kwargs) -> None:
        try:
            super().write(*args, **kwargs)
        except:
            log.warning("while writing on fifo, an error occured, ignored.")

        self.check_buffer()

    def reset(self) -> None:
        super().read(
            samples=max(self.samples - self.SAMPLES_PER_FRAME, 0), partial=True
        )



import array
import ctypes
import ctypes.util
import os
import sys

from ..config import Config

_library = None

c_int_pointer = ctypes.POINTER(ctypes.c_int)
c_int16_pointer = ctypes.POINTER(ctypes.c_int16)
c_float_pointer = ctypes.POINTER(ctypes.c_float)


class EncoderStructure(ctypes.Structure):
    pass


EncoderStructurePointer = ctypes.POINTER(EncoderStructure)


def errorLt(result, func, args):
    if result < 0:
        raise Exception
    return result


def errorNe(result, func, args):
    ret = args[-1]._obj
    if ret.value != 0:
        raise Exception
    return result


exported_functions = {
    "opus_strerror": ([ctypes.c_int], ctypes.c_char_p, None),
    "opus_encoder_get_size": ([ctypes.c_int], ctypes.c_int, None),
    "opus_encoder_create": (
        [ctypes.c_int, ctypes.c_int, ctypes.c_int, c_int_pointer],
        EncoderStructurePointer,
        errorNe,
    ),
    "opus_encode": (
        [
            EncoderStructurePointer,
            c_int16_pointer,
            ctypes.c_int,
            ctypes.c_char_p,
            ctypes.c_int32,
        ],
        ctypes.c_int32,
        errorLt,
    ),
    "opus_encoder_ctl": (None, ctypes.c_int32, errorLt),
    "opus_encoder_destroy": ([EncoderStructurePointer], None, None),
}


def loadOpus(name: str) -> bool:
    global _library
    _library = loadLibopus(name)

    return _library is not None


def loadDefaultOpus() -> bool:
    global _library
    try:
        if sys.platform == "win32":
            architecture: str = "x64" if sys.maxsize > 32 ** 2 else "x86"
            directory: str = os.path.dirname(os.path.abspath(__file__))
            _library = loadLibopus(
                os.path.join(directory, "bin", f"libopus-0.{architecture}.dll")
            )
        else:
            _library = loadLibopus(ctypes.util.find_library("opus"))
    except:
        _library = None

    return _library is not None


def loadLibopus(name: str) -> ctypes.cdll:
    library = ctypes.cdll.LoadLibrary(name)

    for FuncName, Options in exported_functions.items():
        Func = getattr(library, FuncName)

        try:
            if Options[1]:
                Func.argtypes = Options[0]
            Func.restype = Options[1]
        except KeyError:
            pass

        try:
            if Options[2]:
                Func.errcheck = Options[2]
        except KeyError:
            pass

    return library


def isLoaded() -> bool:
    global _library
    return _library is not None


ENCODER_CTL = {
    "OK": 0,
    "APPLICATION_AUDIO": 2049,
    "APPLICATION_VOIP": 2048,
    "APPLICATION_LOWDELAY": 2051,
    "CTL_SET_BITRATE": 4002,
    "CTL_SET_BANDWIDTH": 4008,
    "CTL_SET_FEC": 4012,
    "CTL_SET_PLP": 4014,
    "CTL_SET_SIGNAL": 4024,
}

BAND_CTL = {
    "narrow": 1101,
    "medium": 1102,
    "wide": 1103,
    "superwide": 1104,
    "full": 1105,
}

SIGNAL_CTL = {
    "auto": -1000,
    "voice": 3001,
    "music": 3002,
}


class Encoder:
    def __init__(self, application: int = ENCODER_CTL["APPLICATION_AUDIO"]) -> None:
        self.application = application

        self.SAMPLING_RATE: int = Config.SAMPLING_RATE
        self.CHANNELS: int = Config.CHANNELS
        self.FRAME_LENGTH: int = Config.FRAME_LENGTH
        self.SAMPLE_SIZE: int = Config.SAMPLE_SIZE
        self.SAMPLES_PER_FRAME: int = Config.SAMPLES_PER_FRAME
        self.EXPECTED_PACKETLOSS: float = Config.EXPECTED_PACKETLOSS
        self.BITRATE: int = Config.BITRATE

        if not isLoaded() and not loadDefaultOpus():
            raise ValueError(
                "Cannot load libopus, please install `libopus-dev` if you are using linux."
            )

        self.state = self.createState()
        self.setBitrate(self.BITRATE)
        self.setFec(True)
        self.setExpectedPacketLoss(self.EXPECTED_PACKETLOSS)
        self.setBandwidth("full")
        self.setSignalType("music")

    def createState(self) -> int:
        ret = ctypes.c_int()
        return _library.opus_encoder_create(
            self.SAMPLING_RATE, self.CHANNELS, self.application, ctypes.byref(ret)
        )

    def __del__(self) -> None:
        if hasattr(self, "state"):
            _library.opus_encoder_destroy(self.state)
            self.state = None

    def setBitrate(self, kbps: int) -> None:
        kbps = min(512, max(16, int(kbps)))

        _library.opus_encoder_ctl(
            self.state, ENCODER_CTL["CTL_SET_BITRATE"], kbps * 1024
        )

    def setBandwidth(self, req: str) -> None:
        if not req in BAND_CTL:
            raise KeyError

        _library.opus_encoder_ctl(
            self.state, ENCODER_CTL["CTL_SET_BANDWIDTH"], BAND_CTL[req]
        )

    def setSignalType(self, req: str) -> None:
        if not req in SIGNAL_CTL:
            raise KeyError

        _library.opus_encoder_ctl(
            self.state, ENCODER_CTL["CTL_SET_SIGNAL"], SIGNAL_CTL[req]
        )

    def setFec(self, enabled: bool = True) -> None:
        _library.opus_encoder_ctl(
            self.state, ENCODER_CTL["CTL_SET_FEC"], 1 if enabled else 0
        )

    def setExpectedPacketLoss(self, percentage: float) -> None:
        _library.opus_encoder_ctl(self.state, ENCODER_CTL["CTL_SET_PLP"], percentage)

    def encode(self, Pcm: bytes, FrameSize: int = None) -> bytes:
        if not FrameSize:
            FrameSize = self.SAMPLES_PER_FRAME

        max_length = len(Pcm)
        Pcm = ctypes.cast(Pcm, c_int16_pointer)
        Data = (ctypes.c_char * max_length)()

        Encoded = _library.opus_encode(self.state, Pcm, FrameSize, Data, max_length)

        return array.array("b", Data[:Encoded]).tobytes()



import asyncio
import collections
import functools
import traceback
from typing import Callable, Coroutine


class EventDispatcher:
    def __init__(self, loop: asyncio.AbstractEventLoop = None) -> None:
        self._Events = collections.defaultdict(list)
        self._listeners = collections.defaultdict(list)
        self._Any = []

        self.loop = loop or asyncio.get_event_loop()

    def on(self, event: str, func: Callable):
        self._Events[event].append(func)
        return self

    def off(self, event: str, func: Callable):
        self._Events[event].remove(func)
        return self

    def onAny(self, func: Callable):
        self._Any.append(func)
        return self

    def offAny(self, func: Callable):
        self._Any.remove(func)
        return self

    def dispatch(self, event_: str, *args, **kwargs) -> None:
        if self._Any:
            for func in self._Any:
                try:
                    if asyncio.iscoroutinefunction(func):
                        self.loop.create_task(func(event_, *args, **kwargs))
                    else:
                        self.loop.call_soon_threadsafe(
                            functools.partial(func, event_, *args, **kwargs)
                        )
                except:
                    traceback.print_exc()

        listeners = self._listeners.get(event_)
        if listeners:
            for (Future, condition) in listeners:
                if not (Future.done() or Future.cancelled()):
                    try:
                        result = condition(*args, **kwargs)
                    except Exception as exception:
                        Future.set_exception(exception)
                        result = True
                    else:
                        if result:
                            Future.set_result(
                                args if len(args) > 1 else (args[0] if args else None)
                            )

                    if result:
                        listeners.remove((Future, condition))

            if not listeners:
                self._listeners.pop(event_)

        if not event_ in self._Events:
            return

        for func in self._Events[event_]:
            try:
                if asyncio.iscoroutinefunction(func):
                    asyncio.run_coroutine_threadsafe(func(*args, **kwargs), self.loop)
                else:
                    self.loop.call_soon_threadsafe(
                        functools.partial(func, *args, **kwargs)
                    )
            except:
                traceback.print_exc()

    def event(self, event: str):
        def wrapper(func):
            self.on(event, func)

        return wrapper

    def once(self, event: str, func: Callable):
        def wrapper(*args, **kwargs):
            self.off(event, func)
            return event(*args, **kwargs)

        return self.on(event, wrapper)

    def wait_for(
        self, event: str, condition: Callable = None, timeout: float = None
    ) -> Coroutine:
        Future = self.loop.create_future()

        if not condition:
            condition = lambda *_: True

        self._listeners[event].append((Future, condition))

        return asyncio.wait_for(Future, timeout)



from typing import Union

import aiohttp

from ...config import Config
from .melon import resolve as melon_resolve
from .spotify import resolve as spotify_resolve


async def resolve(
    query: str, connector: aiohttp.TCPConnector = None
) -> Union[str, list]:
    if "melon" in Config.ENABLED_EXT_RESOLVER:
        query = await melon_resolve(query, connector)
    if "spotify" in Config.ENABLED_EXT_RESOLVER:
        query = await spotify_resolve(query)

    return query



import asyncio
import audioop
import functools
import threading
import traceback
from typing import Coroutine

import av

from ..config import Config
from ..natives import AudioFifo, AudioFilter
from ..utils.threadLock import withLock

AVOption = {
    "err_detect": "ignore_err",
    "reconnect": "1",
    "reconnect_streamed": "1",
    "reconnect_delay_max": "5",
}


class PyAVSource:
    def __init__(self, Source: str) -> None:
        self.loop = asyncio.get_event_loop()

        self.Source: str = Source
        self.AVOption: dict = AVOption
        self.Container: av.StreamContainer = None
        self.selectAudioStream = self.FrameGenerator = None

        self._end = threading.Event()
        self._haveToReloadResampler = threading.Event()
        self._waitforread = threading.Lock()
        self._loading = threading.Lock()
        self._seeking = threading.Lock()
        self.BufferLoader: Loader = None

        self.AudioFifo = AudioFifo()
        self._duration: float = None
        self._position: float = 0.0
        self._volume: float = 1.0
        self._filter: dict = {}
        self.stopped: bool = False

    def __del__(self):
        self.cleanup()

    @property
    def volume(self) -> float:
        return self._volume

    @volume.setter
    def volume(self, value: float) -> None:
        self._volume = max(value, 0.0)

    @property
    def filter(self) -> dict:
        return self._filter

    @filter.setter
    def filter(self, value: dict) -> None:
        self._filter = value

    @property
    def duration(self) -> float:
        return self._duration

    @property
    def position(self) -> float:
        return round(
            self._position
            - (
                self.AudioFifo.samples
                / Config.SAMPLES_PER_FRAME
                / 50
                * (float(self.filter["atempo"]) if "atempo" in self.filter else 1.0)
            ),
            2,
        )

    def read(self) -> bytes:
        if not self.BufferLoader:
            self.start()

        if not self.AudioFifo:
            return

        Data = self.AudioFifo.read()
        if not Data and self._loading.locked():
            while self._loading.locked():
                if not self._waitforread.locked():
                    self._waitforread.acquire()
                self._waitforread.acquire()

                Data = self.AudioFifo.read()
                if Data:
                    break

        if Data and self.volume != 1.0:
            Data = audioop.mul(Data, 2, min(self._volume, 2.0))

        return Data

    def _seek(self, offset: float, *args, **kwargs) -> None:
        with withLock(self._seeking):
            if not self.Container:
                if not self._loading.locked():
                    self.Container = av.open(
                        self.Source.Source, options=self.Source.AVOption
                    )
                else:
                    while not self.Container:
                        pass

            kwargs["any_frame"] = True

            self.Container.seek(round(max(offset, 1) * 1000000), *args, **kwargs)
            self.reload()

    def seek(self, offset: float, *args, **kwargs) -> Coroutine:
        return self.loop.run_in_executor(
            None, functools.partial(self._seek, offset, *args, **kwargs)
        )

    def reload(self) -> None:
        self._haveToReloadResampler.set()

        if not self._loading.locked():
            if self._end.is_set():
                self._end.clear()
            self.start()

    def start(self) -> None:
        self.BufferLoader = Loader(self)
        self.BufferLoader.start()

    def stop(self) -> bool:
        self.stopped = True
        return self.stopped

    def is_opus(self) -> bool:
        return False

    def cleanup(self) -> None:
        self._end.set()
        if self.AudioFifo and not self.AudioFifo.haveToFillBuffer.is_set():
            self.AudioFifo.haveToFillBuffer.clear()
        self.AudioFifo = None


class Loader(threading.Thread):
    def __init__(self, AudioSource: PyAVSource) -> None:
        threading.Thread.__init__(self)
        self.daemon = True

        self.Source = AudioSource

        self.Resampler = None
        self.Filter = {}
        self.FilterGraph = None

    def _do_run(self) -> None:
        with withLock(self.Source._loading):
            if not self.Source.Container:
                self.Source.Container = av.open(
                    self.Source.Source, options=self.Source.AVOption
                )
            self.Source._duration = round(self.Source.Container.duration / 1000000, 2)

            self.Source.selectAudioStream = self.Source.Container.streams.audio[0]
            self.Source.FrameGenerator = self.Source.Container.decode(
                self.Source.selectAudioStream
            )

            while not self.Source._end.is_set():
                if self.Source.filter != self.Filter:
                    self.Filter = self.Source.filter

                    if self.Source.filter:
                        self.FilterGraph = AudioFilter()
                        self.FilterGraph.selectAudioStream = (
                            self.Source.selectAudioStream
                        )
                        self.FilterGraph.setFilters(self.Filter)
                    else:
                        self.FilterGraph = None

                if not self.Resampler or self.Source._haveToReloadResampler.is_set():
                    self.Resampler = av.AudioResampler(
                        format=av.AudioFormat("s16").packed, layout="stereo", rate=48000
                    )
                    self.Source._haveToReloadResampler.clear()

                _seek_locked = False
                if self.Source._seeking.locked():
                    self.Source._seeking.acquire()
                    _seek_locked = True

                Frame = next(self.Source.FrameGenerator, None)

                if _seek_locked:
                    self.Source._seeking.release()
                    self.Source.AudioFifo.reset()

                if not Frame:
                    self.Source.stop()
                    break

                _current_position = float(Frame.pts * Frame.time_base)

                if self.FilterGraph:
                    self.FilterGraph.push(Frame)
                    Frame = self.FilterGraph.pull()

                    if not Frame:
                        continue

                Frame.pts = None
                try:
                    Frame = self.Resampler.resample(Frame)
                except ValueError:
                    self.Source._haveToReloadResampler.set()
                    continue

                if self.Source.AudioFifo:
                    if not self.Source.AudioFifo.haveToFillBuffer.is_set():
                        self.Source.AudioFifo.haveToFillBuffer.wait()

                    self.Source.AudioFifo.write(Frame)

                self.Source._position = _current_position

                if self.Source._waitforread.locked():
                    self.Source._waitforread.release()

    def run(self) -> None:
        try:
            self._do_run()
        except:
            traceback.print_exc()
        finally:
            if self.Source.Container:
                self.Source.Container.close()
                self.Source.Container = None

            self.Source.stop()



import re
from xml.etree import ElementTree

import aiohttp
from markdownify import markdownify

from .abc import SubtitleFormat


class smi(SubtitleFormat):
    BODY_REGEX = re.compile(r"<body>([\w|\W|\s]+)</body>", flags=re.I)
    SYNC_REGEX = re.compile(r"<Sync Start=([0-9]+)>([\s|\w|\W]+)", flags=re.I)

    def __init__(self, Body: str) -> None:
        super().__init__()

        self._TextElements = []

        for Text in [
            Item.strip()
            for Item in self.BODY_REGEX.search(Body).group(1).splitlines()
            if Item.strip()
        ]:
            if Text.lower().startswith("<sync"):
                Match = self.SYNC_REGEX.match(Text)

                Start, Element = round(int(Match.group(1)) / 1000, 2), Match.group(2)

                if self._TextElements:
                    self._TextElements[-1]["end"] = Start - 1
                    self._TextElements[-1]["duration"] = (
                        self._TextElements[-1]["end"] - self._TextElements[-1]["start"]
                    )

                self._TextElements.append({"start": Start, "text": Element})
            elif self._TextElements:
                self._TextElements[-1]["text"] += "\n" + Text

        self.TextElements = {
            TextElement["start"]: dict(
                TextElement,
                markdown=markdownify(TextElement["text"]).strip(),
                end=TextElement.get("end") or TextElement["start"] + 5,
                duration=TextElement.get("duration") or 5,
            )
            for TextElement in self._TextElements
            if markdownify(TextElement["text"]).strip()
        }

        self.duration = sorted(
            [TextElement["end"] for TextElement in self.TextElements.values()]
        )[-1]

    @classmethod
    async def load(cls, URL: str):
        async with aiohttp.ClientSession() as session:
            async with session.get(URL) as session:
                Data = await session.text()

        # Blocking is suspicious
        return cls(Data)


class srv1(SubtitleFormat):
    def __init__(self, Tree: ElementTree):
        super().__init__()

        self.Tree = Tree

        self.TextElements = {
            float(TextElement.attrib["start"]): {
                "start": float(TextElement.attrib["start"]),
                "duration": float(TextElement.attrib["dur"]),
                "end": round(
                    float(TextElement.attrib["start"])
                    + float(TextElement.attrib["dur"]),
                    2,
                ),
                "text": TextElement.text,
                "markdown": markdownify(TextElement.text),
            }
            for TextElement in self.Tree.findall("text")
            if "dur" in TextElement.attrib
        }

        self.duration = sorted(
            [TextElement["end"] for TextElement in self.TextElements.values()]
        )[-1]

    @classmethod
    async def load(cls, URL: str):
        async with aiohttp.ClientSession() as session:
            async with session.get(URL) as session:
                Data = await session.text()

        Tree = ElementTree.fromstring(Data)
        return cls(Tree)



import asyncio
import copy
import ipaddress
from typing import Coroutine, Union

from youtube_dlc import YoutubeDL as YoutubeDLClient

from ..errors import NoSearchResults

YTDLOption = {
    "format": "(bestaudio[ext=opus]/bestaudio/best)[protocol!=http_dash_segments]",
    "nocheckcertificate": True,
    "no_warnings": True,
    "default_search": "auto",
    "source_address": "0.0.0.0",
    "skip_download": True,
    "writesubtitles": True,
    "noplaylist": True,
}


def _extract(
    query: str,
    address: Union[ipaddress.IPv4Address, ipaddress.IPv6Address] = None,
    video: bool = False,
) -> dict:
    option = copy.copy(YTDLOption)

    if video:
        option["format"] = "(best)[protocol!=http_dash_segments]"

    if address:
        option["source_address"] = str(address)

    YoutubeDL = YoutubeDLClient(option)
    Data = YoutubeDL.extract_info(query, download=False)

    if not Data:
        raise NoSearchResults

    if "entries" in Data:
        if len(Data["entries"]) == 1:
            return Data["entries"][0]

        return Data["entries"]

    if not Data:
        raise NoSearchResults

    return Data


def _clear_cache() -> None:
    option = {
        "ignoreerrors": True,
        "no_warnings": True,
    }

    YoutubeDL = YoutubeDLClient(option)
    YoutubeDL.cache.remove()


def extract(
    query: str,
    address: Union[ipaddress.IPv4Address, ipaddress.IPv6Address] = None,
    video: bool = False,
    loop: asyncio.AbstractEventLoop = None,
) -> Coroutine:
    if not loop:
        loop = asyncio.get_event_loop()

    return loop.run_in_executor(None, _extract, query, address, video)


def clear_cache(loop: asyncio.AbstractEventLoop = None) -> Coroutine:
    if not loop:
        loop = asyncio.get_event_loop()

    return loop.run_in_executor(None, _clear_cache)



import abc


class SubtitleFormat(abc.ABC):
    def __init__(self) -> None:
        self.time = 0.0
        self.current = None

    def seek(self, time: float) -> str:
        self.time = float(time)

        if (
            not self.current
            or not self.current["start"] <= self.time < self.current["end"]
        ):
            Elements = [
                TextElement
                for Start, TextElement in self.TextElements.items()
                if Start <= self.time < TextElement["end"]
            ]
            self.current = Elements[-1] if Elements else None

        return self.lyrics

    def __dict__(self) -> dict:
        return self.TextElements

    @property
    def lyrics(self) -> str:
        if self.current:
            return self.current
        return None

    @property
    def is_done(self) -> bool:
        return self.time >= self.duration

    @classmethod
    @abc.abstractmethod
    async def load(cls, URL: str):
        raise NotImplementedError



import asyncio

from .errors import VoiceClientNotFound
from .event import DiscordEvent
from .source import AudioData
from .utils import EventDispatcher
from .voice_client import VoiceClient


class ClientManager:
    def __init__(self, **kwargs) -> None:
        self.loop = asyncio.get_event_loop()

        self.id = kwargs.get("user_id")
        self.session_id = kwargs.get("session_id")

        self.voiceClients = {}

        self.dispatcher = EventDispatcher()
        self.event = self.dispatcher.event
        self.discordEvent = DiscordEvent(self)

    def __del__(self) -> None:
        for voiceClient in self.voiceClients.values():
            self.loop.call_soon_threadsafe(voiceClient.__del__)

    def __repr__(self) -> str:
        return f"<ClientManager id={self.id} session_id='{self.session_id}' voiceClients={len(self.voiceClients)}>"

    def discordDispatch(self, data: dict) -> None:
        Event, Data = data["t"], data["d"]

        return self.discordEvent.dispatch(Event, data=Data)

    def getVC(self, guildID: int, safe: bool = False) -> VoiceClient:
        if not int(guildID) in self.voiceClients and not safe:
            raise VoiceClientNotFound

        return self.voiceClients.get(int(guildID))

    def delVC(self, guildID: int) -> None:
        self.getVC(guildID).__del__()

    async def getSource(self, Query: str) -> AudioData:
        return await AudioData.create(Query)

    def putSource(self, guildID: int, *args, **kwargs) -> int:
        return self.getVC(guildID).putSource(*args, **kwargs)

    async def loadSource(self, guildID: int, *args, **kwargs) -> AudioData:
        return await self.getVC(guildID).loadSource(*args, **kwargs)

    def skip(self, guildID: int, offset: int) -> None:
        return self.getVC(guildID).skip(offset)

    async def seek(self, guildID: int, offset: int) -> None:
        return await self.getVC(guildID).seek(offset)

    def setVolume(self, guildID: int, value: float) -> float:
        self.getVC(guildID).volume = value

        return self.getVC(guildID).volume

    def setCrossfade(self, guildID: int, value: float) -> float:
        self.getVC(guildID).crossfade = value

        return self.getVC(guildID).crossfade

    def setGapless(self, guildID: int, value: bool) -> bool:
        self.getVC(guildID).gapless = value

        return self.getVC(guildID).gapless

    def setAutoplay(self, guildID: int, value: bool) -> bool:
        self.getVC(guildID).autoplay = value

        return self.getVC(guildID).autoplay



from typing import Any, Dict, List

import aiohttp
import yarl
import youtube_dlc

from ..config import Config
from ..errors import Forbidden, TooManyRequests
from ..extractor import extract
from ..extractor.youtube_dl import clear_cache
from .AudioSource import AudioSource


class AudioData:
    def __init__(self, data: dict) -> None:
        self.id = data["id"]
        self.title = data.get("title")

        if data.get("_type") == "url" and data.get("ie_key") == "Youtube":
            self.webpage_url = f"https://www.youtube.com/watch?v={self.id}"
            self.thumbnail = f"https://i.ytimg.com/vi/{self.id}/hqdefault.jpg"
            self.stream_url = None
        else:
            self.webpage_url = data.get("webpage_url")
            self.thumbnail = data.get("thumbnail")
            self.stream_url = data.get("url")

        self.duration = data.get("duration")
        self.is_live = data.get("is_live", False)

        self.uploader = data.get("uploader")
        self.description = data.get("description")

        def getUsableSubtitle(Data: Dict[str, Any]) -> Dict[str, Any]:
            Usable: List[Any] = list(
                filter(lambda Subtitle: Subtitle.get("ext") == "srv1", Data)
            )

            if Usable:
                return Usable[0]["url"]

        self.subtitles = (
            dict(
                map(
                    lambda Data: (
                        Data[0],
                        getUsableSubtitle(Data[1]),
                    ),
                    data["subtitles"].items(),
                )
            )
            if "subtitles" in data
            else {}
        )

        self.chapters = data.get("chapters", {})

        self.related: bool = False

        self._source = None

    def __dict__(self) -> dict:
        return {
            "_type": "AudioData",
            "id": self.id,
            "title": self.title,
            "webpage_url": self.webpage_url,
            "thumbnail": self.thumbnail,
            "url": self.stream_url,
            "duration": self.duration,
            "is_live": self.is_live,
            "uploader": self.uploader,
            "description": self.description,
            "subtitles": self.subtitles,
            "chapters": self.chapters,
            "related": self.related,
        }

    def __repr__(self) -> str:
        return f"<AudioData id={self.id} title='{self.title}' duration={self.duration} address='{self.address}'>"

    @classmethod
    async def create(cls, query: str):
        cls.address = Config.RoutePlanner.get() if Config.RoutePlanner else None

        try:
            Data = await extract(query, address=cls.address)
        except youtube_dlc.utils.DownloadError as exc:
            if Config.RoutePlanner and exc.exc_info[1].status == 429:
                Config.RoutePlanner.mark_failed_address(cls.address)

                return await cls.create(query)

            raise exc

        if isinstance(Data, list):
            return [cls(Item) for Item in Data]

        return cls(Data)

    async def gather(self):
        try:
            Data = await extract(self.webpage_url, address=self.address)
        except youtube_dlc.utils.DownloadError as exc:
            if Config.RoutePlanner and exc.exc_info[1].status == 429:
                Config.RoutePlanner.mark_failed_address(self.address)
                self.address = Config.RoutePlanner.get()

                return await self.gather()

            raise exc

        self.__init__(Data)

        return self

    async def source(
        self, *args, _retry: int = 0, _limited: bool = False, **kwargs
    ) -> AudioSource:
        if not self.stream_url or _limited:
            await self.gather()

        if not self._source:
            async with aiohttp.ClientSession() as session:
                URL = yarl.URL(self.stream_url, encoded=True)
                async with session.get(URL) as resp:
                    Status = resp.status

            if Status == 403:
                if _retry == 0:
                    return await self.source(
                        *args, _retry=_retry + 1, _limited=True, **kwargs
                    )
                elif _retry == 1:
                    await clear_cache()
                    return await self.source(
                        *args, _retry=_retry + 1, _limited=True, **kwargs
                    )

                raise Forbidden
            if Status == 429:
                if Config.RoutePlanner:
                    Config.RoutePlanner.mark_failed_address(self.address)
                    self.address = Config.RoutePlanner.get()

                    return await self.source(*args, _limited=True, **kwargs)

                raise TooManyRequests

            self._source = AudioSource(self.stream_url, AudioData=self, **kwargs)

        return self._source
